from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from paystack.models import TransactionLog
from paystack.paystack_urls import (
    PAYSTACK_CHARGE_AUTHORIZATION_URL,
    PAYSTACK_INITIALIZE_TRANSACTION_URL,
    PAYSTACK_VERIFY_TRANSACTION_URL,
)

from .base_api_service import BaseAPIService


class TransactionService(BaseAPIService):
    def _create_transaction_object(self, transaction_data):
        user_id = transaction_data["metadata"]["user_id"]
        user = self.get_user(user_id)

        TransactionLog.objects.create(
            user=user,
            charge_type="GATEWAY PURCHASE",
            amount=transaction_data["amount"],
            currency=transaction_data["currency"],
            txRef=transaction_data["reference"],
            payment_date_time=transaction_data["paid_at"],
            status=transaction_data["status"],
        )

    def log_transaction(
        self, transaction_data
    ):  # transaction will be logged in the webhook
        self._create_transaction_object(transaction_data)

    def _validate_initiate_payload(self, payload: dict) -> None:
        """
        check that payload has all the required params
        """
        required = ["email", "amount"]

        for i in required:
            try:
                payload[i]
            except KeyError:
                raise ValidationError(f"{i} must be provided")

        self.validate_amount(payload["amount"])

    def _validate_charge_payload(self, payload: dict) -> None:
        """
        check that payload has all the required params
        """
        required = ["email", "amount", "authorization_code"]

        for i in required:
            try:
                payload[i]
            except KeyError:
                raise ValidationError(f"{i} must be provided")

        self.validate_amount(payload["amount"])

    def initialize_payment(self, payload: dict) -> Response:
        url = PAYSTACK_INITIALIZE_TRANSACTION_URL

        self._validate_initiate_payload(payload)
        return self.make_request("POST", url, payload)

    def verify_payment(self, transaction_ref: str) -> Response:
        url = PAYSTACK_VERIFY_TRANSACTION_URL.format(transaction_ref)

        return self.make_request("GET", url)

    def recurrent_charge(self, payload: dict) -> Response:
        self._validate_charge_payload(payload)

        url = PAYSTACK_CHARGE_AUTHORIZATION_URL

        return self.make_request("POST", url, payload)



from paystack.models import PayStackCustomer

from .base_api_service import BaseAPIService


class CustomerService(BaseAPIService):
    def _create_customer_object(self, user, customer_data, authorization_data):
        defaults = {
            "user": user,
            "email": customer_data["email"],
            "authorization_code": authorization_data["authorization_code"],
            "card_type": authorization_data["card_type"],
            "last4": authorization_data["last4"],
            "exp_month": authorization_data["exp_month"],
            "exp_year": authorization_data["exp_year"],
            "bin": authorization_data["bin"],
            "bank": authorization_data["bank"],
            "account_name": authorization_data["account_name"],
        }

        PayStackCustomer.objects.update_or_create(**defaults, defaults=defaults)

    def log_customer(self, webhook_data) -> None:
        user_id = webhook_data["metadata"]["user_id"]
        user = self.get_user(user_id)

        customer_data = webhook_data["customer"]
        authorization_data = webhook_data["authorization"]

        self._create_customer_object(user, customer_data, authorization_data)



import pytest

from tests.mixins import RequestMixin, URLsMixin

pytestmark = pytest.mark.django_db


@pytest.mark.usefixtures(
    "valid_transaction_payload",
    "invalid_transaction_payload",
    "transaction_reference",
    "valid_charge_payload",
    "invalid_charge_payload",
)
class TestTransactionEndpoints(URLsMixin, RequestMixin):
    def test_initiate_transaction(self, valid_transaction_payload):
        response = self.send_request(
            request_method="POST",
            request_url=self.initiate_transaction_url(),
            payload=valid_transaction_payload,
        )

        assert response.status_code == 200

    def test_initiate_transaction_fails(self, invalid_transaction_payload):
        response = self.send_request(
            request_method="POST",
            request_url=self.initiate_transaction_url(),
            payload=invalid_transaction_payload,
        )

        assert response.status_code != 200
        assert response.status_code == 400

    def test_verify_transaction(self, transaction_reference):
        response = self.send_request(
            request_method="GET",
            request_url=self.verify_transaction_url(transaction_reference),
        )

        assert response.status_code == 200

    def test_verify_transaction_fails(self):
        response = self.send_request(
            request_method="GET",
            request_url=self.verify_transaction_url(""),
        )

        assert response.status_code != 200
        assert response.status_code == 400

    def test_charge_customer(self, valid_charge_payload):
        response = self.send_request(
            request_method="POST",
            request_url=self.charge_customer_url(),
            payload=valid_charge_payload,
        )

        assert response.status_code == 200

    def test_charge_customer_fails(self, invalid_charge_payload):
        response = self.send_request(
            request_method="POST",
            request_url=self.charge_customer_url(),
            payload=invalid_charge_payload,
        )

        assert response.status_code != 200
        assert response.status_code == 400

    def test_list_transaction(self):
        response = self.send_request(
            request_method="GET", request_url=self.all_transactions_url()
        )

        assert response.status_code == 200



PAYSTACK_BASE_URL = "https://api.paystack.co"

TRANSACTION_URL = PAYSTACK_BASE_URL + "/transaction"
PAYSTACK_INITIALIZE_TRANSACTION_URL = TRANSACTION_URL + "/initialize"
PAYSTACK_VERIFY_TRANSACTION_URL = TRANSACTION_URL + "/verify/{0}"  # insert tran. ref.

PAYSTACK_CHARGE_AUTHORIZATION_URL = TRANSACTION_URL + "/charge_authorization"



from django.conf import settings
from rest_framework import status
from rest_framework.authentication import TokenAuthentication
from rest_framework.response import Response


def return_okay_response(data=None, status=status.HTTP_200_OK) -> Response:
    response = {"status": "success", "result": data}
    return Response(response, status=status)


def return_bad_response(data, status=status.HTTP_400_BAD_REQUEST) -> Response:
    response = {"status": "Failure", "result": data}
    return Response(response, status=status)


def get_authentication_class():

    try:
        return getattr(settings, "REST_FRAMEWORK")["DEFAULT_AUTHENTICATION_CLASSES"]

    except Exception:
        return (TokenAuthentication,)



from django.urls import reverse


class URLsMixin(object):
    def initiate_transaction_url(self):
        return reverse("transaction-initiate")

    def verify_transaction_url(self, trans_ref):
        return reverse("transaction-verify") + f"?transaction_ref={trans_ref}"

    def charge_customer_url(self):
        return reverse("transaction-charge-customer")

    def transaction_url(self, transaction_id):
        return reverse("transaction-detail")

    def all_transactions_url(self):
        return reverse("transaction-list")

    def webhook_handler_url(self):
        return reverse("webhook-handler")

    def get_customer_url(self, user_id):
        return reverse("customer-detail")

    def all_customers_url(self):
        return reverse("customer-list")



# import pytest

# from paystack.models import PayStackCustomer
# from paystack.services import CustomerService
# from tests.mock_data import webhook_data

# pytestmark = pytest.mark.django_db


# class TestCustomerService(CustomerService):
#     @pytest.mark.django_db
#     def test_log_customer(self):
#         self.log_customer(webhook_data)

#         assert PayStackCustomer.objects.all().count() > 0



import hashlib
import hmac

from django.conf import settings
from rest_framework.exceptions import ValidationError

from .customer_service import CustomerService
from .transaction_service import TransactionService


class WebhookService(object):
    def __init__(self, request) -> None:
        self.request = request

    def webhook_handler(self):
        try:
            secret = getattr(settings, "PAYSTACK_PRIVATE_KEY")
        except Exception as e:  # If user hasn't declared variable
            raise ValidationError(e)

        webhook_data = self.request.data
        hash = hmac.new(secret, webhook_data, digestmod=hashlib.sha512).hexdigest()

        if hash != self.request.headers["x-paystack-signature"]:
            raise ValidationError("MAC authentication failed")

        if webhook_data["event"] == "charge.success":
            paystack_service = TransactionService()
            paystack_service.log_transaction(webhook_data["data"])

            customer_service = CustomerService()
            customer_service.log_customer(webhook_data["data"])

        return webhook_data



from django.contrib import admin

from paystack.models import PayStackCustomer, TransactionLog


class PayStackCustomerAdmin(admin.ModelAdmin):
    model = PayStackCustomer
    list_display = ["user", "email", "authorization_code"]


class TransactionLogAdmin(admin.ModelAdmin):
    model = TransactionLog
    list_display = ["uuid", "user", "charge_type", "amount"]


admin.site.register(PayStackCustomer, PayStackCustomerAdmin)
admin.site.register(TransactionLog, TransactionLogAdmin)



from rest_framework.views import APIView

from paystack.services import WebhookService
from paystack.utils import return_okay_response


class WebhookFacadeView(APIView):
    """
    Exsits for extensibility reasons. Users might want to capture
    the data returned from Paystack and do some stuff with it.
    E.g  retrieve the user tied to the
    payment(usually passed as a meta data in this package)
    and clear the user's cart or create an order for that user.
    """

    authentication_classes = []
    permission_classes = []

    def post(self, request):
        webhook_service = WebhookService(request)
        return (
            webhook_service.webhook_handler()
        )  # This returns raw JSON data from Paystack


class WebhookView(WebhookFacadeView):
    def post(self, request):
        webhook_data = super().post(request)

        return_okay_response(webhook_data)  # Return instance of JsonResponse



import pytest

from tests.mixins import RequestMixin, URLsMixin

pytestmark = pytest.mark.django_db


class TestCustomerEndpoints(URLsMixin, RequestMixin):
    def test_list_customers(self):
        response = self.send_request(
            request_method="GET", request_url=self.all_customers_url()
        )

        assert response.status_code == 200



from rest_framework import viewsets
from rest_framework.decorators import action

from paystack.models import TransactionLog
from paystack.serializers import PaymentSerializer
from paystack.services import TransactionService
from paystack.utils import get_authentication_class, return_okay_response


class TransactionViewSet(viewsets.ModelViewSet):
    queryset = TransactionLog.objects.all()
    serializer_class = PaymentSerializer
    http_method_names = ["get", "post"]
    authentication_classes = get_authentication_class()
    lookup_field = "uuid"

    @action(detail=False, methods=["post"])
    def initiate(self, request):
        """
        This is used to charge customers that had already been charged in
        past.
        Expects the payload in the format below:

        {
            "email": "string",
            "amount": float/int,
            "metadata": dict/json, --Optional
        }
        """
        payload = request.data

        if "metadata" in payload:
            payload["metadata"]["user_id"] = request.user.id
        else:
            payload["metadata"] = {"user_id": request.user.id}

        transaction_service = TransactionService()
        initiated_transaction = transaction_service.initialize_payment(payload)

        return return_okay_response(initiated_transaction)

    @action(detail=False, methods=["get"])
    def verify(self, request):
        transaction_ref = request.query_params.get("transaction_ref")

        paystack_service = TransactionService()
        verified_transaction = paystack_service.verify_payment(transaction_ref)

        return return_okay_response(verified_transaction)

    @action(detail=False, methods=["post"], url_path="charge-customer")
    def charge_customer(self, request):
        """
        This is used to charge customers that had already been charged in
        past.
        Expects the payload in the format below:

        {
            "email": "string",
            "amount": float/int,
            "authorization_code": "string",
        }

        """
        payload = request.data

        transaction_service = TransactionService()
        charge = transaction_service.recurrent_charge(payload)

        return return_okay_response(charge)



import json

import pytest
from django.contrib.auth import get_user_model
from django.utils.encoding import force_str
from rest_framework.authtoken.models import Token
from rest_framework.test import APIClient

pytestmark = pytest.mark.django_db

User = get_user_model()


@pytest.mark.usefixtures("client")
class RequestMixin(object):
    @pytest.mark.django_db
    def _client(self):
        user, _ = User.objects.get_or_create(
            email="admin@gmail.com", password="Gr3@t!2021"
        )

        token, _ = Token.objects.get_or_create(user=user)

        client = APIClient()
        client.credentials(HTTP_AUTHORIZATION="Token " + token.key)

        return client

    def send_request(self, **kwargs):
        request_method = kwargs.get("request_method").lower()
        request_url = kwargs.get("request_url")

        client = self._client()
        if request_method == "post":
            response = client.post(request_url, data=kwargs["payload"], format="json")

        if request_method == "get":
            response = client.get(request_url)

        return response



import uuid

from django.contrib.auth import get_user_model
from django.db import models

User = get_user_model()


class PayStackCustomer(models.Model):
    """
    for charging a customer's card again using authorization code
    for transfers too
    """

    user = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)

    # authorization creds for recurring charges
    email = models.CharField(blank=True, null=True, max_length=100)
    authorization_code = models.CharField(blank=True, null=True, max_length=100)
    card_type = models.CharField(blank=True, null=True, max_length=10)
    last4 = models.CharField(blank=True, null=True, max_length=4)
    exp_month = models.CharField(blank=True, null=True, max_length=10)
    exp_year = models.CharField(blank=True, null=True, max_length=10)
    bin = models.CharField(blank=True, null=True, max_length=10)
    bank = models.CharField(blank=True, null=True, max_length=100)
    account_name = models.CharField(blank=True, null=True, max_length=100)


class TransactionLog(models.Model):

    GATEWAY_PURCHASE = "GATEWAY_PURCHASE"
    TRANSFER = "TRANSFER"

    uuid = models.UUIDField(unique=True, default=uuid.uuid4)
    user = models.ForeignKey(User, on_delete=models.SET_NULL, null=True)
    charge_type = models.CharField(
        max_length=30,
        choices=(
            (GATEWAY_PURCHASE, "Gateway Purchase"),
            (TRANSFER, "Transfer"),
        ),
    )
    amount = models.FloatField(max_length=19)
    currency = models.CharField(max_length=5)
    txRef = models.CharField(max_length=100, null=True, blank=True)

    payment_date_time = models.DateTimeField(max_length=100, null=True, blank=True)
    status = models.CharField(max_length=50, null=True, blank=True)



from rest_framework import serializers

from paystack.models import TransactionLog


class PaymentSerializer(serializers.ModelSerializer):
    class Meta:
        model = TransactionLog
        fields = "__all__"



import json

import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from rest_framework.exceptions import ValidationError

User = get_user_model()


class BaseAPIService(object):  # Not to be instantiated directly
    def get_user(self, user_id):
        user = User.objects.get(id=user_id)
        return user

    def make_request(self, method, url, payload=None):

        try:
            PAYSTACK_PRIVATE_KEY = getattr(settings, "PAYSTACK_PRIVATE_KEY")
        except Exception as e:  # If ser hasn't declared variable
            raise ValidationError(e)

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer { PAYSTACK_PRIVATE_KEY }",
        }

        response = requests.request(
            method, url, data=json.dumps(payload), headers=headers
        )

        if response.status_code != 200:
            if response.text:
                raise ValidationError(response.text)
            else:
                raise ValidationError(
                    f"paystack failed with error code: {response.status_code}"
                )

        data_json_str = json.dumps(json.loads(response.text))
        # convert json str to json object
        result = json.loads(data_json_str)

        return result

    def validate_amount(self, amount):

        if isinstance(amount, int) or isinstance(amount, float):
            if amount < 0:
                raise ValidationError("Negative amount is not allowed")
            return amount * 100  # in kobo
        else:
            raise ValidationError("Amount must be a number")

    def validate_email(self, email):
        if not email:
            raise ValidationError("Customer Email is required")



import pytest
from rest_framework.exceptions import ValidationError

from paystack.models import TransactionLog
from paystack.services import TransactionService
from tests.mock_data import webhook_data

pytestmark = pytest.mark.django_db


class TestTransactionService(TransactionService):
    @pytest.mark.django_db
    def test_log_transaction(self):
        self.log_transaction(webhook_data)

        assert TransactionLog.objects.all().count() > 0

    def test_validate_initiate_transaction_payload(self, invalid_transaction_payload):
        with pytest.raises(ValidationError):
            self._validate_initiate_payload(invalid_transaction_payload)

    def test_validate_charge_payload(self, invalid_charge_payload):
        with pytest.raises(ValidationError):
            self._validate_charge_payload(invalid_charge_payload)



from rest_framework import serializers

from paystack.models import PayStackCustomer


class CustomerSerializer(serializers.ModelSerializer):
    class Meta:
        model = PayStackCustomer
        fields = "__all__"



# Generated by Django 2.2 on 2021-12-12 11:26

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="TransactionLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4, unique=True)),
                (
                    "charge_type",
                    models.CharField(
                        choices=[
                            ("GATEWAY_PURCHASE", "Gateway Purchase"),
                            ("TRANSFER", "Transfer"),
                        ],
                        max_length=30,
                    ),
                ),
                ("amount", models.FloatField(max_length=19)),
                ("currency", models.CharField(max_length=5)),
                ("txRef", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "payment_date_time",
                    models.DateTimeField(blank=True, max_length=100, null=True),
                ),
                ("status", models.CharField(blank=True, max_length=50, null=True)),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="PayStackCustomer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("email", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "authorization_code",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                ("card_type", models.CharField(blank=True, max_length=10, null=True)),
                ("last4", models.CharField(blank=True, max_length=4, null=True)),
                ("exp_month", models.CharField(blank=True, max_length=10, null=True)),
                ("exp_year", models.CharField(blank=True, max_length=10, null=True)),
                ("bin", models.CharField(blank=True, max_length=10, null=True)),
                ("bank", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "account_name",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                (
                    "user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



from django.urls import include, path
from rest_framework import routers

from paystack.views import PaystackCustomerViewSet, TransactionViewSet, WebhookView

router = routers.DefaultRouter()
router.register("transaction", TransactionViewSet, basename="transaction")
router.register("paystack-customer", PaystackCustomerViewSet, basename="customer")

urlpatterns = [
    path("", include(router.urls)),
    path("webook-handler", WebhookView.as_view(), name="webhook-handler"),
]



from rest_framework import viewsets

from paystack.models import PayStackCustomer
from paystack.serializers import CustomerSerializer
from paystack.utils import get_authentication_class


class PaystackCustomerViewSet(viewsets.ModelViewSet):
    queryset = PayStackCustomer.objects.all()
    serializer_class = CustomerSerializer
    http_method_names = ["get"]
    authentication_classes = get_authentication_class()
    lookup_field = "user__id"
